package com.oas.modbusutilities.configuration;

import java.text.MessageFormat;
import java.util.logging.Logger;

import javax.baja.control.BNumericWritable;
import javax.baja.driver.point.BPointDeviceExt;
import javax.baja.status.BStatus;
import javax.baja.sys.BComponent;
import javax.baja.sys.BDouble;
import javax.baja.sys.BLink;
import javax.baja.sys.Context;
import javax.baja.sys.Property;

import com.oas.modbusutilities.enums.BOasAutobauding;
import com.oas.modbusutilities.enums.BOasBautrate;
import com.oas.modbusutilities.enums.BOasModuleCmd;
import com.oas.modbusutilities.enums.BOasStatusLed;
import com.oas.modbusutilities.utils.AsyncLinkUtils;
import com.oas.modbusutilities.utils.LinkUtils;
import com.tridium.modbusAsync.BModbusAsyncDevice;
import com.tridium.modbusCore.client.datatypes.BCommStatus;
import com.tridium.modbusCore.client.datatypes.BModbusClientPresetRegister;
import com.tridium.modbusCore.client.datatypes.BModbusClientPresetRegisters;
import com.tridium.modbusCore.client.enums.BCommStatusEnum;


@SuppressWarnings({
  "squid:S2629", // Invoke LOGs conditionally will bloat up the code
})
public class AsyncBasicConfiguration {


  private static final Logger LOG = Logger.getLogger("oasModbusUtilities");
  private static final String LOG_PREFIX = AsyncBasicConfiguration.class.getSimpleName();


  private AsyncBasicConfiguration() { }

  public static void doWriteModuleCmd(BComponent reg, BComponent preset, BOasModuleCmd cmd) {
    if (reg instanceof BNumericWritable) {
      BNumericWritable moduleCmd = (BNumericWritable)reg;
      moduleCmd.doSet(BDouble.make(cmd.getOrdinal()));
    }

    if (preset instanceof BModbusClientPresetRegisters) {

      BModbusClientPresetRegisters moduleCmdPreset = (BModbusClientPresetRegisters)preset;
      moduleCmdPreset.doWrite();
    }
    else {
      LOG.warning(MessageFormat.format("{0}.doWriteModuleCmd(): Object was not of type BModbusClientPresetRegisters",
                                    LOG_PREFIX));
    }
  }


  public static void writeAutobinding(BComponent reg, BOasAutobauding value) {
    if (reg instanceof BModbusClientPresetRegisters) {
      LOG.info(MessageFormat.format("{0}.updateAutobinding(): upate to {1}",
                                    LOG_PREFIX, value.getOrdinal()));
      BModbusClientPresetRegisters autobauding = (BModbusClientPresetRegisters)reg;
      autobauding.doWrite();
    }
    else {
      LOG.warning(MessageFormat.format("{0}.writeAutobinding(): Object was not of type BModbusClientPresetRegisters",
                                    LOG_PREFIX));
    }
  }


  public static void writeBaudreate(BComponent reg, BOasBautrate value) {
    if (reg instanceof BModbusClientPresetRegisters) {
      LOG.info(MessageFormat.format("{0}.updateBaudreate(): upate to {1}",
                                    LOG_PREFIX, value.getOrdinal()));
      BModbusClientPresetRegisters baudrate = (BModbusClientPresetRegisters)reg;
      baudrate.doWrite();
    }
    else {
      LOG.warning(MessageFormat.format("{0}.writeBaudreate(): Object was not of type BModbusClientPresetRegisters",
                                    LOG_PREFIX));
    }
  }


  public static void writeTimeout(BComponent reg) {
    if (reg instanceof BModbusClientPresetRegisters) {
      BModbusClientPresetRegisters timeoutSecs = (BModbusClientPresetRegisters)reg;
      timeoutSecs.doWrite();
    }
    else {
      LOG.warning(MessageFormat.format("{0}.writeTimeout(): Object was not of type BModbusClientPresetRegisters",
                                    LOG_PREFIX));
    }

  }


  public static void writeLedStatus(BComponent reg, BComponent pres, BOasStatusLed statusLedOneValue, BOasStatusLed statusLedTwoValue) {
    if (reg instanceof BNumericWritable) {
      BNumericWritable statusLed = (BNumericWritable)reg;
      Double ledValue = statusLed.getOut().getNumeric();
      int newLedValue = BOasStatusLed.DEFAULT.getOrdinal();
      int oldLedValue = ledValue.intValue();
      if (statusLedTwoValue != BOasStatusLed.off) {
        oldLedValue -= statusLedTwoValue.getOrdinal();
      }
      if(oldLedValue == BOasStatusLed.OFF) {
        // always set status led one color
        newLedValue = statusLedOneValue.getOrdinal();
      }
      // if status led two is set, add it to the output value
      newLedValue += statusLedTwoValue.getOrdinal();

      statusLed.doSet(BDouble.make(newLedValue));
    }
    if (pres instanceof BModbusClientPresetRegisters) {
      BModbusClientPresetRegisters preset = (BModbusClientPresetRegisters)pres;
      preset.doWrite();
    }
  }


  public static void createLinksToRegister(BModbusAsyncDevice module, BAbstractConfiguration config) {

    LOG.fine(MessageFormat.format("{0}.createLinksToRegister({1}) {2}.{3}: create Links to Config Component",
                                    LOG_PREFIX, config.getDeviceDisplayName(null), config.getDeviceName(), config.getName()));
    AsyncConfiguration async = null;
    if (config instanceof AsyncConfiguration) {
      async = (AsyncConfiguration) config;
    }
    else {
      LOG.warning(MessageFormat.format(
          "{0}.createLinksToRegister({1}): {2} does not impl. AsyncConfiguration Interface! Linking could not be done",
          LOG_PREFIX, config.getDeviceDisplayName(null), config.getName()));
      return;
    }


    BComponent autobaudingRegister = config.getAutobaudingRegister();
    if (null != autobaudingRegister) {

      Property srcProp = autobaudingRegister.getProperty("out");
      BLink autobaudingLink = LinkUtils.createLink(autobaudingRegister, srcProp, config, BAbstractConfiguration.activeAutobauding);
      if (autobaudingLink == null) {
        LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not link to Autobauding",
                                            LOG_PREFIX, config.getDeviceDisplayName(null)));
      }
      else {
        config.add("autobaudingLink", autobaudingLink);
      }
    }
    else {
      LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not find Register for Autobauding",
                                       LOG_PREFIX, config.getDeviceDisplayName(null)));
    }


    BComponent autobauingPreset = async.getAutobaudingPreset();
    if (null != autobauingPreset) {
      if (autobauingPreset instanceof BModbusClientPresetRegisters) {
        BModbusClientPresetRegisters preset = (BModbusClientPresetRegisters)autobauingPreset;

        AsyncLinkUtils.linkToPresets(preset, (BComponent)config, BAbstractConfiguration.setPointAutobauding.getName());

      }
      else {
        LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not create link to Autobauding preset {2}",
                                          LOG_PREFIX, config.getDeviceDisplayName(null), autobauingPreset.getName()));
      }
    }
    else {

      LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not find preset for Autobauding",
                                        LOG_PREFIX, config.getDeviceDisplayName(null)));
    }



    BComponent baudrateRegister = config.getBaudrateRegister();
    if (null != baudrateRegister) {

      Property srcProp = baudrateRegister.getProperty("out");
      BLink link = LinkUtils.createLink(baudrateRegister, srcProp, config, BAbstractConfiguration.activeBaudrate);
      if (link == null) {
        LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not link to Baudrate",
                                            LOG_PREFIX, config.getDeviceDisplayName(null)));
      }
      else {
        config.add("baudrateLink", link);
      }
    }
    else {
      LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not find Register for Baudrate",
                                        LOG_PREFIX, config.getDeviceDisplayName(null)));
    }

    BComponent baudratePreset = async.getBaudratePreset();
    if (null != baudratePreset) {
      if (baudratePreset instanceof BModbusClientPresetRegisters) {
        BModbusClientPresetRegisters preset = (BModbusClientPresetRegisters)baudratePreset;

        AsyncLinkUtils.linkToPresets(preset, (BComponent)config, BAbstractConfiguration.setPointBaudrate.getName());

      }
      else {
        LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not create link to Baudrate preset {2}",
                                          LOG_PREFIX, config.getDeviceDisplayName(null), baudratePreset.getName()));
      }
    }
    else {
      LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not find preset for Baudrate",
                                        LOG_PREFIX, config.getDeviceDisplayName(null)));
    }

    BComponent moduleTypeRegister = config.getModuleTypeRegister();
    if (null != moduleTypeRegister) {

      Property srcProp = moduleTypeRegister.getProperty("out");
      BLink link = LinkUtils.createLink(moduleTypeRegister, srcProp, config, BAbstractConfiguration.activeModuleType);
      if (link == null) {
        LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not link to ModuleType",
                                            LOG_PREFIX, config.getDeviceDisplayName(null)));
      }
      else config.add("moduleTypeLink", link);
    }
    else {
      LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not find Register for ModuleType",
                                        LOG_PREFIX, config.getDeviceDisplayName(null)));
    }

    BComponent timeoutSecsRegister = config.getTimeoutRegister();
    if (null != timeoutSecsRegister) {

      Property srcProp = timeoutSecsRegister.getProperty("out");
      BLink link = LinkUtils.createLink(timeoutSecsRegister, srcProp, config, BAbstractConfiguration.activeTimeout);
      if (link == null) {
        LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not link to TimeoutSecs",
                                            LOG_PREFIX, config.getDeviceDisplayName(null)));
      }
      else {
        config.add("timeoutSecsLink", link);
      }
    }
    else {
      LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not find Register for TimeoutSecs",
                                      LOG_PREFIX, config.getDeviceDisplayName(null)));
    }

    BComponent timeoutSecsPreset = async.getTimeoutPreset();
    if (null != timeoutSecsPreset) {
      if (timeoutSecsPreset instanceof BModbusClientPresetRegisters) {
        BModbusClientPresetRegisters preset = (BModbusClientPresetRegisters)timeoutSecsPreset;

        AsyncLinkUtils.linkToPresets(preset, (BComponent)config, BAbstractConfiguration.setPointTimeout.getName());

      }
      else {
        LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not create link to timeoutSecs preset {2}",
                                          LOG_PREFIX, config.getDeviceDisplayName(null), timeoutSecsPreset.getName()));
      }
    }
    else {
      LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not find preset for timeoutSecs",
                                        LOG_PREFIX, config.getDeviceDisplayName(null)));
    }

    BComponent firmwareVersionRegister =config.getFirmwareVersionRegister();
    if (null != firmwareVersionRegister) {

      Property srcProp = firmwareVersionRegister.getProperty("out");
      BLink link = LinkUtils.createLink(firmwareVersionRegister, srcProp, config, BAbstractConfiguration.displayFirmware);
      if (link == null) {
        LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not link to FirmwareVersion",
                                            LOG_PREFIX, config.getDeviceDisplayName(null)));
      }
      else {
        config.add("firmwareVersionLink", link);
      }
    }
    else {
      LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not find Register for FirmwareVersion",
                                      LOG_PREFIX, config.getDeviceDisplayName(null)));
    }

    if (null != module) {

      Property srcProp = BModbusAsyncDevice.deviceAddress;
      BLink link = LinkUtils.createLink(module, srcProp, config, BAbstractConfiguration.displayAddress);
      if (link == null) {
        LOG.warning(MessageFormat.format("{0}.createLinksToRegister({1}) could not link to Address",
                                            LOG_PREFIX, config.getDeviceDisplayName(null)));
      }
      else {
        config.add("addressLink", link);
      }
    }
  }

  public static BStatus getDeviceStatus(BModbusAsyncDevice device) {

      if (null != device) {
        return device.getStatus();
      }
      return null;
  }


  public static boolean isDeviceEnabeld(BModbusAsyncDevice device) {

    if (null != device) {
      return device.getEnabled();
    }
    return false;
  }


  public static boolean isDeviceOk(BModbusAsyncDevice device) {
    return device.getStatus().isOk();
  }


  public static String getDeviceName(BModbusAsyncDevice device) {

    if (null != device) {

      return device.getName();
    }

    return null;

  }



  public static String getDeviceDisplayName(BModbusAsyncDevice device, Context ctx) {

    if (null != device) {

      return device.getDisplayName(ctx);
    }

    return null;

  }


  public static BComponent[] getRegister(BModbusAsyncDevice device) {

    if (null != device) {
      BPointDeviceExt pointExt = device.getPointDeviceExt();
      if (null != pointExt) {

        return pointExt.getChildComponents();
      }
    }

    return new BComponent[] {};

  }


  public static void checkPresetWriteStatus(BModbusClientPresetRegisters presets, BAbstractConfiguration config) {
    BComponent[] presetEntries = presets.getChildComponents();
    BConfigurationStatus status = config.getStatus();

    for (BComponent entry : presetEntries) {
      if (entry instanceof BModbusClientPresetRegister) {
        BModbusClientPresetRegister pEntry = (BModbusClientPresetRegister)entry;
        BCommStatus writeStatus = pEntry.getWriteStatus();

        switch (writeStatus.getErrorCode()) {
          case BCommStatusEnum.ILLEGAL_FUNCTION:
            // change device status to configFailure
            LOG.warning(MessageFormat.format(
                                "{0}.checkPresetWriteStatus() detected ILLEGAL_FUNCTION write Status at {1}.{2}",
                                LOG_PREFIX, presets.getName(), entry.getName()));
            config.setStatus(BConfigurationStatus.makeConfigFailure(status, true));
          break;
          case BCommStatusEnum.ILLEGAL_DATA_ADDRESS:
            // change device status to configFailure
            LOG.warning(MessageFormat.format(
                                "{0}.checkPresetWriteStatus() detected ILLEGAL_DATA_ADDRESS write Status at {1}.{2}",
                                LOG_PREFIX, presets.getName(), entry.getName()));

            config.setStatus(BConfigurationStatus.makeConfigFailure(status, true));

          break;
          case BCommStatusEnum.ILLEGAL_DATA_VALUE:
            // change device status to configFailure
            LOG.warning(MessageFormat.format(
                                "{0}.checkPresetWriteStatus() detected ILLEGAL_DATA_VALUE write Status at {1}.{2}",
                                LOG_PREFIX, presets.getName(), entry.getName()));
            config.setStatus(BConfigurationStatus.makeConfigFailure(status, true));
          break;
          case BCommStatusEnum.SLAVE_DEVICE_FAILURE:
            // log that write status is SLAVE_DEVICE_FAILURE
            LOG.severe(MessageFormat.format("{0}.checkPresetWriteStatus() detected {1} Status on {2}.{3}",
                                           LOG_PREFIX, BCommStatusEnum.slaveDeviceFailure.getTag(),
                                           presets.getName(), entry.getName()));
          break;
          case BCommStatusEnum.SLAVE_DEVICE_BUSY:
            // log that write status is SLAVE_DEVICE_BUSY
            LOG.severe(MessageFormat.format("{0}.checkPresetWriteStatus() detected {1} Status on {2}.{3}",
                                           LOG_PREFIX, BCommStatusEnum.slaveDeviceFailure.getTag(),
                                           presets.getName(), entry.getName()));
          break;
          case BCommStatusEnum.NEGATIVE_ACKNOWLEDGE:
            // change device status to configFailure
            LOG.warning(MessageFormat.format(
                                "{0}.checkPresetWriteStatus() detected NEGATIVE_ACKNOWLEDGE write Status at {1}.{2}",
                                LOG_PREFIX, presets.getName(), entry.getName()));

            config.setStatus(BConfigurationStatus.makeConfigFailure(status, true));

          break;
          default:
            LOG.finest(MessageFormat.format(
                     "{0}.checkPresetWriteStatus() no status detected which is of concern at {1}.{2}",
                     LOG_PREFIX, presets.getName(), entry.getName()));
             // remove configFailure from device status
            if (status.isConfigFailure()) {
              config.setStatus(BConfigurationStatus.make(status.getBits() & ~BConfigurationStatus.CONFIG_FAILURE));
            }
        }
      }
    }
  }

}
