define(function(require) {
    'use strict';


    var baja = require('baja!');
    var ConfigurationStatus = require('baja!oasModbusUtilitiesCore:ConfigurationStatus');
    var types = require('baja!control:PriorityLevel');
    var lexicons = require('lex!oasModbusUtilitiesCore');
    var Promise = require('Promise');
    var rangeUtils = require('nmodule/webEditors/rc/fe/baja/util/rangeUtils');

    var lex = lexicons[0];
    var getEnumRangeDisplay = rangeUtils.getEnumRangeDisplay;


    var lexKeys = [
        "ConfigStatus.modified",
        "ConfigStatus.wrongModuleType",
        "ConfigStatus.autobaudingModified",
        "ConfigStatus.baudrateModified",
        "ConfigStatus.timeoutModified",
        "ConfigStatus.offline"
    ];

    var ACTIVE_LEVEL = 'activeLevel';


    /**
    * @private
    * @exports nmodule/webEditors/rc/fe/baja/util/statusUtils
    */
    var exports = {};


    function statusBitsToString(bits) {
        var statuses, i;

        if (bits === 0) {
          statuses = [lex.get('ConfigStatus.ok')];
        } else {
          statuses = [];

          for (i = 0; i < lexKeys.length; i++) {
            if (bits & (1 << i)) {
              statuses.push(lex.get(lexKeys[i]));
            }
          }
        }

        return '{' + statuses.join(',') + '}';
      }

    /**
    * `DynamicEnum`s don't have display tags, but `FrozenEnum`s do. This function
    * converts a `DynamicEnum` backed by a `PriorityLevel` into the corresponding
    * display tag, pulling it from the `control` lexicon in the case of
    * `fallback`. Other Simple values just get toStringed.
    *
    * @private
    * @inner
    * @param {baja.Simple} activeLevel
    * @returns {String}
    * @see BPriorityLevel.getDisplayTag()
    */
    function activeLevelToString(activeLevel) {
        return '@ ' + (toPriorityLevelString(activeLevel) || activeLevel);
    }


    //TODO: this needs to be a PriorityLevel type extension (NCCB-24257)
    function toPriorityLevelString(priorityLevel) {
        if (!baja.hasType(priorityLevel, 'baja:DynamicEnum')) {
            return null;
        }

        var frozen = priorityLevel.getRange().getFrozenType();
        if (frozen && frozen.is('control:PriorityLevel')) {
            var frozenEnum = baja.$('control:PriorityLevel').get(priorityLevel.getOrdinal());
            if (frozenEnum.getTag() === 'fallback') {
                return controlLex.get('def');
            } else {
                return frozenEnum.getDisplayTag();
            }
        }
    }

    function facetsToString(facets) {
        var keys = facets.getKeys();
        var key;
        var value;
        var a = [];
        var i;

        for (i = 0; i < keys.length; i++) {
            key = keys[i];
            value = facets.get(key);

            if (key === ACTIVE_LEVEL) {
                a.push(activeLevelToString(value));
            } else if (value === true) {
                a.push(key);
            } else {
                a.push(key + '=' + value);
            }
        }

        return a.join(' ');
    }

    function simpleToString(simple, obj) {
        switch (String(simple.getType())) {

            case 'baja:Boolean':
                var text = obj[simple ? 'trueText' : 'falseText'];
                return text ? baja.Format.format(text) : bajaLex.get(String(simple));

            case 'baja:Double':
                return simple.toString(obj);

            case 'baja:DynamicEnum':
                return getEnumRangeDisplay(simple.getOrdinal(), obj.range || simple.getRange());

            case 'baja:String':
                return simple;

            default:
                return String(simple);
        }
    }


    /**
     * Converts a `oasModbusUtilitiesCore:ConfigurationStatus` to a display string in a similar way to `ConfigurationStatus`.
     *
     * @param {oasModbusUtilitiesCore.ConfigurationStatus|Number} status the status to encode (a bit string in
     * the form of a number will be converted to a `oasModbusUtilitiesCore:ConfigurationStatus`)
     * @returns {Promise} promise to be resolved with the toString-ed
     * status, or rejected if the input was invalid
     */
    exports.statusToString = function statusToString(status) {
        if (typeof status === 'number') {
            status = ConfigurationStatus.make(status);
        }

        if (!baja.hasType(status, 'oasModbusUtilitiesCore:ConfigurationStatus')) {
            return Promise.reject(new Error('oasModbusUtilitiesCore:ConfigurationStatus or number required'));
        }

        var str = statusBitsToString(status.getBits());
        var facets = status.getFacets();

        if (facets && (facets !== baja.Facets.DEFAULT)) {
            str += ' ' + facetsToString(facets);
        }

        return Promise.resolve(str);
    };

    /**
     * Resolve a human-readable string value for a `StatusValue`.
     *
     * Supports `trueText` and `falseText` for booleans, `range` for enums,
     * and all properties (`precision` etc.) that are supported by `numberUtils`.
     *
     * @param {baja.Struct} statusValue a `baja:StatusValue`
     * @param {Object} [config] properties to use to format the `value` property
     * @returns {*}
     */
    exports.statusValueToString = function (statusValue, config) {
        if (!baja.hasType(statusValue, 'baja:StatusValue')) {
            return Promise.reject(new Error('baja:StatusValue required'));
        }

        var status = statusValue.get('status');
        var value = statusValue.get('value');

        return Promise.join(
            status.isNull() ? '-' : simpleToString(value, config || {}),
            exports.statusToString(status)
        )
        .spread(function (valueString, statusString) {
            return valueString + ' ' + statusString;
        });
  };

  return (exports);
});
