define(function(require) {
    'use strict';

    var baja = require("baja!");
    var Widget = require('bajaux/Widget');
    var subscriberMixIn = require('bajaux/mixin/subscriberMixIn');

    var BajaUtilities = require('nmodule/oasModbusUtilitiesCore/rc/script/BajaUtilities');
    var ModuleTypeInfo = require("nmodule/oasModbusUtilitiesCore/rc/script/moduleTypeInfo");

    var Header = require("nmodule/oasModbusUtilitiesCore/rc/script/Header");

    var OverviewTempl = require('hbs!nmodule/oasModbusUtilitiesCore/rc/template/overview');
    var TableRowTempl = require('hbs!nmodule/oasModbusUtilitiesCore/rc/template/overviewTableRow');

    require('css!nmodule/oasModbusUtilitiesCore/rc/style/overview');


    var lexicons = require('lex!oasModbusUtilitiesCore');
    var lex = lexicons[0];

    var ModuleOverview = function (params) {

        Widget.apply(this, arguments);

        subscriberMixIn(this);

        this.$subscribers = [];

        this.$widget = null;

        this.$header = new Header();
    };

    ModuleOverview.prototype = Object.create(Widget.prototype);
    ModuleOverview.prototype.constructor = ModuleOverview;


    ModuleOverview.prototype.doInitialize = function(dom) {
        dom.html(OverviewTempl({
            header: this.$header.toSafeString(),
            station_name: "",
            address_label: lex.get('Overview.table.address'),
            description_label: lex.get('Overview.table.description'),
            moduletype_label: lex.get('Overview.table.modultype'),
            config_label: lex.get('Overview.table.config'),
            status_label: lex.get('Overview.table.status')
        }));
    };

    ModuleOverview.prototype.doLoad = function(comp) {

        var jq = this.jq();
        var that = this;

        this.$widget = comp;


        var homeOrd = new baja.Ord.make(this.$widget.getHomeOrd());
        var logoOrd = new baja.Ord.make(this.$widget.getLogoOrd());

        this.$header.setHomeLink(homeOrd.toUri());
        this.$header.setBrandLogo(logoOrd.toUri());

        var serialPortConfig = comp.getSerialPortConfig();


        this.getStationname().then(function(name) {
            jq.find('.station-name').text(name);
        });

        jq.find('.modbus-config').text(comp.getName() + " - " + serialPortConfig);

        var modules = this.getModules(comp);
        this.updateTable(modules);



        this.getSubscriber().attach('added', function(prop) {
            that.updateTable([prop.$val]);
        });

        this.getSubscriber().attach('removed', function(prop) {
            var module = prop.$val;
            // console.log(module);
            // console.log(" with Handle: " + module.getHandle() + " was removed.");

            if (module.getType().is('modbusAsync:ModbusAsyncDevice')) {
                console.log("Async Device was removed");
                jq.find('#' + module.getHandle()).remove();
            }
        });

    };

    ModuleOverview.prototype.updateTable = function(modules) {

        var that = this;
        var jq = this.jq();
        var moduleTable = jq.find('#module-table');


        modules.forEach(function(module) {

            // console.log("handle " + module.getName());

            module.loadSlots().then(function() {


                // console.log(module);
                var moduleConfig = that.getModuleConfig(module);
                var moduleHandle = module.getHandle();
                var modileSlotPath = module.getSlotPath();

                moduleConfig.loadSlots().then(function() {

                    var typOk = moduleConfig.getTypOk().getValue();

                    var moduleType = ModuleTypeInfo.searchModuleType(moduleConfig.getActiveModuleType());
                    var info = {
                        module_handle: moduleHandle,
                        address: module.getDeviceAddress(),
                        description: module.getName(),
                        module_status: BajaUtilities.evaluateMeterStatus(module.getStatus()),
                        type: moduleType,
                        show_status: !typOk,
                        config_status: "is " + moduleType,
                        overview_link: modileSlotPath
                    };
                    moduleTable.append(TableRowTempl(info));
                });

                var moduleSubscriber = new baja.Subscriber();
                moduleSubscriber.subscribe(module);
                moduleSubscriber.attach('changed', function(prop){

                    if (prop.getName() === 'deviceAddress') {
                        that.setRowModuleAddress(jq, moduleHandle, prop.$val);
                    }
                    else if (prop.getName() === 'status') {
                        that.setRowModuleStatus(jq, moduleHandle, prop.$val);
                    }
                });


                var moduleConfigSubscriber = new baja.Subscriber();
                moduleConfigSubscriber.subscribe(moduleConfig);
                moduleConfigSubscriber.attach("changed", function(prop) {
                    if (prop.getName() === 'activeModuleType') {
                        that.setRowModuleType(jq, moduleHandle, prop.$val);
                    }
                    else if (prop.getName() === 'typOk') {
                        that.setRowModuleTypeStatus(jq, moduleHandle, prop.$val, moduleConfig.getActiveModuleType());
                    }
                });

                jq.on('click', 'div#' + moduleHandle + " div.overview-link", function () {

                    var elem = jq.find(this);


                    var url = "";
                    var ordStr = 'station:|' + elem.data('ord');
                    var viewStr = elem.data('view');
                    var ord = baja.Ord.make(ordStr + viewStr);

                    console.log(ord);

                    if (window.niagara.env.type === "wb") {
                      url = ord.toString();
                    } else {
                      url = ord.toUri();
                    }

                    window.niagara.env.hyperlink(url);
                    return false;
                  });


                that.$subscribers.push(moduleSubscriber);
                that.$subscribers.push(moduleConfigSubscriber);

            });

        });
    };

    ModuleOverview.prototype.setRowModuleAddress = function(jq, select, value) {
        var row = jq.find("#" + select);
        row.find("div.overview-address").text(value);
    };

    ModuleOverview.prototype.setRowModuleStatus = function(jq, select, value) {
        var row = jq.find("#" + select);
        var status = BajaUtilities.evaluateMeterStatus(value);
        row.find("div.overview-description").removeClass(BajaUtilities.allStatusClasses()).addClass(status);
    };

    ModuleOverview.prototype.setRowModuleType = function(jq, select, value) {
        var row = jq.find("#" + select);
        row.find('.overview-moduletype').text(ModuleTypeInfo.searchModuleType(value));
    };


    ModuleOverview.prototype.setRowModuleTypeStatus = function(jq, select, value, type) {
        var row = jq.find("#" + select);
        var configStatus = row.find('.overview-configstatus');
        if(value.getValue()) {
            configStatus.text('');
        }
        else {
            configStatus.text('is ' + ModuleTypeInfo.searchModuleType(type));
        }
    };




    ModuleOverview.prototype.getModuleConfig = function(oasModule) {
        var config = oasModule.getSlots().each(function(slot) {
            // console.log("Slot: " + slot.getName() );
            if (slot.getName() === "configuration") {
                return slot.$val;
            }
        });



        return config;
    };

    ModuleOverview.prototype.getModules = function(component) {

        if (component.getType().is("oasModbusUtilitiesAsync:OasModbusAsyncNetwork")) {
            // console.log("get Modules from Async Driver");
            return component.getSlots().properties().is('modbusAsync:ModbusAsyncDevice').toValueArray();
        }
        else if(component.getType().is("oasModbusUtilitiesPs:OasModbusPsNetwork")){
            //TODO to be implemented
            return null;
        }

        return null;
    };


    ModuleOverview.prototype.getStationname = function() {
        var className = "oasModbusUtilitiesCore:OasServerSideCallHandler";
        var funcName = "getStationName";

        return this.$widget.serverSideCall({
            typeSpec: className,
            methodName: funcName
        }).catch(function(err) {
            console.error("ModuleOverview.getStationame() serverSideCall " +
                        className + "." + funcName + ", err: " + err);
        });
    };

    return ModuleOverview;

});
